split header out to action.h
[model-checker.git] / model.cc
index f1bc320b417e7c7f0938a73d50eff20b2a6b7301..f447c1dfa50e98139b4b1b30a7cb5d9d5a6a5060 100644 (file)
--- a/model.cc
+++ b/model.cc
@@ -1,15 +1,19 @@
 #include <stdio.h>
 
 #include "model.h"
+#include "action.h"
+#include "tree.h"
 #include "schedule.h"
 #include "common.h"
 
+#define INITIAL_THREAD_ID      0
+
 ModelChecker *model;
 
 ModelChecker::ModelChecker()
 {
-       /* First thread created (system_thread) will have id 1 */
-       this->used_thread_id = 0;
+       /* First thread created will have id (INITIAL_THREAD_ID + 1) */
+       this->used_thread_id = INITIAL_THREAD_ID;
        /* Initialize default scheduler */
        this->scheduler = new Scheduler();
 
@@ -41,7 +45,7 @@ void ModelChecker::reset_to_initial_state()
        action_trace = new action_list_t();
        currentNode = rootNode;
        current_action = NULL;
-       used_thread_id = 1; // ?
+       used_thread_id = INITIAL_THREAD_ID;
        /* scheduler reset ? */
 }
 
@@ -117,9 +121,7 @@ bool ModelChecker::next_execution()
 
 ModelAction * ModelChecker::get_last_conflict(ModelAction *act)
 {
-       void *loc = act->get_location();
        action_type type = act->get_type();
-       thread_id_t id = act->get_tid();
 
        switch (type) {
                case THREAD_CREATE:
@@ -135,14 +137,8 @@ ModelAction * ModelChecker::get_last_conflict(ModelAction *act)
        action_list_t::reverse_iterator rit;
        for (rit = action_trace->rbegin(); rit != action_trace->rend(); rit++) {
                ModelAction *prev = *rit;
-               if (prev->get_location() != loc)
-                       continue;
-               if (type == ATOMIC_READ && prev->get_type() != ATOMIC_WRITE)
-                       continue;
-               /* Conflict from the same thread is not really a conflict */
-               if (id == prev->get_tid())
-                       continue;
-               return prev;
+               if (act->is_dependent(prev))
+                       return prev;
        }
        return NULL;
 }
@@ -165,10 +161,12 @@ void ModelChecker::set_backtracking(ModelAction *act)
        if (node->setBacktrack(act->get_tid()) != 0)
                return;
 
-       printf("Setting backtrack: conflict = %d, instead tid = %d\n",
+       DEBUG("Setting backtrack: conflict = %d, instead tid = %d\n",
                        prev->get_tid(), act->get_tid());
-       prev->print();
-       act->print();
+       if (DBG_ENABLED()) {
+               prev->print();
+               act->print();
+       }
 
        Backtrack *back = new Backtrack(prev, action_trace);
        backtrack_list.push_back(back);
@@ -255,6 +253,62 @@ ModelAction::ModelAction(action_type_t type, memory_order order, void *loc, int
        act->value = value;
 }
 
+bool ModelAction::is_read()
+{
+       return type == ATOMIC_READ;
+}
+
+bool ModelAction::is_write()
+{
+       return type == ATOMIC_WRITE;
+}
+
+bool ModelAction::is_acquire()
+{
+       switch (order) {
+       case memory_order_acquire:
+       case memory_order_acq_rel:
+       case memory_order_seq_cst:
+               return true;
+       default:
+               return false;
+       }
+}
+
+bool ModelAction::is_release()
+{
+       switch (order) {
+       case memory_order_release:
+       case memory_order_acq_rel:
+       case memory_order_seq_cst:
+               return true;
+       default:
+               return false;
+       }
+}
+
+bool ModelAction::same_var(ModelAction *act)
+{
+       return location == act->location;
+}
+
+bool ModelAction::same_thread(ModelAction *act)
+{
+       return tid == act->tid;
+}
+
+bool ModelAction::is_dependent(ModelAction *act)
+{
+       if (!is_read() && !is_write())
+               return false;
+       if (!act->is_read() && !act->is_write())
+               return false;
+       if (same_var(act) && !same_thread(act) &&
+                       (is_write() || act->is_write()))
+               return true;
+       return false;
+}
+
 void ModelAction::print(void)
 {
        const char *type_str;